'''Simulator of a currency exchange.'''

import numpy as np
import pandas as pd
from os import path
from os import listdir
from .common import _read_csv


def _read_market(data_path: str) -> pd.core.frame.DataFrame:
    '''Read all csv files with candles inside the directory and compose
    a Pandas DataFrame.

    Args:
        data_path: Where csv files with data are stored.

    Returns:
        A Pandas DataFrame containing all candles for all
            currencies.
    '''
    market = None
    for idx, filename in enumerate(sorted(listdir(data_path))):
        df = _read_csv(path.join(data_path, filename))
        df['currency'] = path.splitext(filename)[0]
        market = pd.concat([market, df])
    return market


def _dict_to_array(dic: dict) -> np.ndarray:
    '''Convert a dictionary to Numpy ndarray.

    Args:
        dic: A dictionary (e.g. balance, portfolio, etc.).

    Returns:
        A Numpy ndarray.
    '''
    return np.array([dic[i] for i in sorted(dic)])


def _array_to_dict(arr: np.ndarray, currencies: tuple) -> dict:
    '''Convert a Numpy ndarray to a dictionary (e.g. balance, portfolio,
    etc.).

    Args:
        arr: A Numpy ndarray.
        currencies: A tuple of currencies.

    Returns:
        A dictionary.
    '''
    return dict(zip(sorted(currencies), arr))


def _candles_to_array(candles: dict) -> np.ndarray:
    '''Convert candle close prices to a dictionary.

    Args:
        candles: A candles dictionary.

    Returns:
        A Numpy ndarray.
    '''
    # Get close prices from all currencies and set 1.0 as a cash price.
    prices = []
    for currency in sorted(list(candles) + ['cash']):
        if currency == 'cash':
            prices.append(1.0)
        else:
            prices.append(candles[currency]['close'])
    return np.array(prices)


def next_step(data_path: str) -> dict:
    '''Go to the next step in timeline, it yelds a one new candle for each
    currency.

    Args:
        data_path: Where csv files with data are stored.

    Returns:
        A dictionary with currencies and their prices.
    '''
    market = _read_market(data_path)
    for step in market.index.unique():
        yield market.loc[step].set_index('currency').to_dict('index')


def capital(candles: dict, balance: dict) -> float:
    '''Returns current capital - sum of all currencies converted to cash
    (without fees).

    Args:
        candles: Current candles as result of next_step function.
        balance: Dictionary of currencies and values representing a current
            balance.

    Returns:
        A capital.
    '''
    capital = 0
    for currency, amount in balance.items():
        if currency == 'cash':
            capital += amount
        else:
            capital += amount * candles[currency]['close']
    return capital


def portfolio(candles: dict, balance: dict) -> dict:
    '''Returns current portfolio - proportion of capital by each currency.

    Args:
        candles: Current candles as result of next_step function.
        balance: Dictionary of currencies and values representing a current
            balance.

    Returns:
        A portfolio.
    '''
    cap = capital(candles, balance)
    portfolio = {}
    for currency, amount in balance.items():
        if currency == 'cash':
            portfolio[currency] = balance[currency] / cap
        else:
            portfolio[currency] = (balance[currency]
                                   * candles[currency]['close'] / cap)
    return portfolio


def buy(candles: dict, balance: dict, currency: str, amount: float,
        fee: float, min_order_size: float) -> dict:
    '''Buy currency.

    Args:
        candles: Current candles.
        balance: Dictionary of currencies and values representing a current
            balance.
        currency: Name of the currency.
        amount: How much units of this currency to buy.
        fee: What part of the trade volume will be paid as fee.
        min_order_size: Minimum trade volume expressed in a base currency
            (cash).

    Returns:
        A new balance after trade.
    '''

    result = balance.copy()
    price = candles[currency]['close']
    without_fee = price * amount
    with_fee = without_fee * (1 + fee)

    if with_fee <= balance['cash'] and without_fee >= min_order_size:
        result['cash'] -= with_fee
        result[currency] += amount

    return result


def sell(candles: dict, balance: dict, currency: str, amount: float,
         fee: float, min_order_size: float) -> dict:
    '''Sell currency.

    Args:
        candles: Current candles.
        balance: Dictionary of currencies and values representing a current
            balance.
        currency: Name of the currency.
        amount: How much units of this currency to sell.
        fee: What part of the trade volume will be paid as fee.
        min_order_size: Minimum trade volume expressed in a base currency
            (cash).

    Returns:
        A new balance after trade.
    '''

    result = balance.copy()
    price = candles[currency]['close']
    without_fee = price * amount
    with_fee = without_fee * (1 - fee)

    if amount <= balance[currency] and without_fee >= min_order_size:
        result['cash'] += with_fee
        result[currency] -= amount

    return result


def make_portfolio(candles: dict, balance: dict, fee: float,
                   target_portfolio: dict) -> dict:
    '''Provide an answer how to make a desired portfolio.

    Args:
        candles: Current candles.
        balance: Dictionary of currencies and values representing a current
            balance.
        fee: What part of the trade volume will be paid as fee.
        target_portfolio: A desired portfolio.

    Returns:
        A dictionary with values of how much of each currency to buy
            (positive number) or sell (negative) in order to achieve a
            desired portfolio.
    '''

    prices = _candles_to_array(candles)
    cur_balance = _dict_to_array(balance)
    cur_capital = capital(candles, balance)
    cur_portfolio = cur_balance * prices / np.sum(cur_balance * prices)
    tar_portfolio = _dict_to_array(target_portfolio.copy())

    # Calculate a capital change after trading.
    cc0 = 1
    cc1 = 1 - 2 * fee + fee ** 2
    while abs(cc1 - cc0) > 1e-10:
        cc0 = cc1
        cc1 = (1 - fee * cur_portfolio[0] - (2 * fee - fee ** 2)
               * np.sum(np.maximum(cur_portfolio[1:]
                        - cc1 * tar_portfolio[1:], 0))) \
            / (1 - fee * tar_portfolio[0])

    # A capital after trade.
    tar_capital = cur_capital * cc1

    tar_balance = tar_capital * tar_portfolio / prices

    return _array_to_dict(tar_balance - cur_balance, balance.keys())



'''Common functions used by several sub-modules. It's for
internal use only.'''

import pandas as pd


def _save_csv(df: pd.core.frame.DataFrame, filepath: str):
    '''Saves Pandas DataFrame to a csv file.

    Args:
        df: Pandas DataFrame.
        filepath: Path of the csv file.
    '''
    df.to_csv(filepath, index=False)


def _read_csv(filepath: str) -> pd.core.frame.DataFrame:
    '''Read csv file into Pandas DataFrame.

    Args:
      filepath: Path to a csv file with data is stored.

    Returns:
        A Pandas DataFrame with candles.
    '''
    return pd.read_csv(filepath)



'''This script downloads candles of selected cryptocurrencies from
poloniex.com and saves them into csv files.

After package installation just run "download_candles" script in your
CLI without any parameters in order to download sample data.
'''

import os
import pandas as pd
from poloniex import Poloniex
from .common import _save_csv


def _request(currency: str, period: int, start: int, end: int) -> list:
    '''Get candles data from Poloniex exchange for a specified currency
    relative to BTC.

    Args:
        currency: Currency which we will request.
        period: Period for one candle in seconds.
        start: Start of the range (UNIX timestamp).
        end: End of the range (UNIX timestamp).

    Returns:
        A list of dictionaries, each of each represents one candle.
    '''
    polo = Poloniex()
    pair = f"BTC_{currency}"
    return polo.returnChartData(pair, period, start=start, end=end)


def _candles_to_df(candles: list) -> pd.core.frame.DataFrame:
    '''Converts list of candles to a Pandas DataFrame.

    Args:
        candles: List of candles.

    Returns Pandas DataFrame filled with candles.
    '''
    return pd.DataFrame(candles, columns=[
            'date', 'high', 'low', 'open', 'close',
            'volume', 'quoteVolume', 'weightedAverage'
        ])


def _main(currencies: list = ['ETH', 'ETC', 'XMR', 'LTC'],
          start: int = 1575158400,
          end: int = 1575244800,
          period: int = 1800,
          data_folder: int = 'sample_data'):
    '''Download candles for specified currencies and range and save them
    to a separate csv files.

    Args:
        currencies: List of currencies to request from Poloniex.
        start: Starting date of the requested range (UNIX timestamp, default
            value is December 1, 2019).
        end: Ending date of the requested range (UNIX timestamp, default
            value is December 2, 2019).
        period: Period for one candle in seconds (Half-hour candles by
            default).
        data_folder: Folder where to save market data.
    '''

    if not os.path.exists(data_folder):
        os.makedirs(data_folder)

    for currency in currencies:
        candles = _request(currency, period, start, end)
        df = _candles_to_df(candles)
        _save_csv(df, f"{data_folder}/{currency}.csv")

    print(f"Market data saved to {data_folder}.")



import codecs
import os.path
from setuptools import setup


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError('Unable to find version string.')


PACKAGE_NAME = 'xchg'

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name=PACKAGE_NAME,
    version=get_version(f"{PACKAGE_NAME}/__init__.py"),
    description='Simulator of a currency exchange.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Sergei Bondarenko',
    author_email='sergei@bondarenko.xyz',
    url=f"https://github.com/sergei-bondarenko/{PACKAGE_NAME}",
    license='Unlicense',
    packages=[PACKAGE_NAME],
    install_requires=[
        'pandas',
        'poloniex',
        'numpy'
    ],
    entry_points={
        'console_scripts':
            [f"download_sample={PACKAGE_NAME}.download_sample:_main"],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)



from .main import next_step
from .main import capital
from .main import portfolio
from .main import buy
from .main import sell
from .main import make_portfolio

__version__ = '1.0.6'
__all__ = [
    'next_step',
    'capital',
    'portfolio',
    'buy',
    'sell',
    'make_portfolio'
]
